<?php

/**
 * @file
 * Sync and management
 */
// ----  Other Functions

/**
 * Generate and return the site callback url.
 * This URL will be called when a document translation is complete, and can be downloaded.
 * Format:  ?doc_id={document_id}&target_code={target_language}&project_id={project_id}&completed={completed}
 * */
function lingotek_notify_url_generate() {
  global $base_url;
  return $base_url . "/" . LINGOTEK_NOTIFY_URL;
}

/**
 * Update the Notify URL (via API)
 */
function lingotek_notify_url_update($show_messages = TRUE) {
  $integration_id = variable_get('lingotek_integration_id', NULL);
  $success = LingotekSync::updateNotifyUrl();
  if ($show_messages) {
    if ($success) {
      if ($success === 'success') {
        drupal_set_message(t("Notification callback URL successfully updated."));
      }
      elseif ($success === 'localhost_url') {
        drupal_set_message(t("Lingotek will not be able to notify Drupal, since the callback URL contains <i>localhost</i>."), 'warning');
      }
    }
    else {
      drupal_set_message(t("Notification callback URL was not successfully updated."), 'error');
    }
  }
  return $success;
}

/**
 * Notify URL Check - call lingotek_notify_url_update when lingotek_notify_url no longer matches lingotek_notify_url_generate()
 */
function lingotek_notify_url_check() {
  if (strcasecmp(variable_get('lingotek_notify_url', ''), lingotek_notify_url_generate()) != 0) {
    lingotek_notify_url_update(FALSE);
  }
}

function lingotek_get_trans_obj($document_id, $document_idx) {
  // Adding a delay in the update.  Without the delay all the different language updates hit at once, causing node lock issues as multiple languages try to update the same node at once.
  $min = 0;
  $max = 3;
  $sleep = rand($min, $max);
  sleep($sleep);
  $trans_obj = NULL;

  // try 3 times if by chance we don't find the document in the database yet..
  $attempts = 0;
  while ($attempts < 3) {
    list($id, $type) = LingotekSync::getEntityIdFromDocId($document_id);
    if (!$id) { // check for the old style document_id
      list($id, $type) = LingotekSync::getEntityIdFromDocId($document_idx);
      if ($id) {
        $document_id = $document_idx;
      }
    }

    if ($id) {
      $entity = lingotek_entity_load_single($type, $id);
      $trans_obj = LingotekEntity::load($entity, $type);
      break;
    }

    if ($trans_obj = LingotekConfigSet::loadByLingotekDocumentId($document_id)) {
      break;
    }

    LingotekLog::info('Did not find doc ID @doc_id yet on attempt #@attempt, retrying...',
        array('@attempt' => $attempts, '@doc_id' => $document_id));
    sleep(2);
    $attempts++;
  }
  if (!$trans_obj) {
    LingotekLog::error('Did not find doc ID @doc_id after all attempts.  Giving up.' , array('@doc_id' => $document_id));
  }
  return $trans_obj;
}

function _lingotek_notification_output($message) {
  return drupal_json_output(array("message" => $message));
}

function _lingotek_notification_handle_target_or_phase($trans_obj, $document_id) {
  $locale = "";
  if (isset($_GET['locale'])) {
    $locale = $_GET['locale']; // Project Callbacks
  } else if (isset($_GET['target_code'])) {
    $locale = $_GET['target_code']; // Integration Callback
  }

  $completed = false;
  if ((isset($_GET['complete']) && $_GET['complete'] != 'false') || (isset($_GET['completed']) && $_GET['completed'] != 'false')){
    $completed = true;
  }
  $project_id = isset($_GET['project_id']) ? $_GET['project_id'] : "";
  $trans_obj->preDownload($locale, $completed);

  $replacements = array(
    '@trans_type' => get_class($trans_obj),
    '@document' => $document_id,
    '@language_code' => $locale,
    '@project_id' => $project_id,
    '@id' => $trans_obj->getId(),
  );

  if ($trans_obj->downloadTriggered($locale)) {
    LingotekLog::trace('Updated local content for <strong>@trans_type</strong> @id based on hit
          from external API for document: @document, language code @language_code, project ID: @project_id', $replacements, 'api');
  }
  else {
    LingotekLog::trace('Unable to update local content for <strong>@trans_type</strong> @id based on hit
          from external API for document: @document, language code @language_code, project ID: @project_id', $replacements, 'api');
  }
  $trans_obj->setStatus(LingotekSync::STATUS_CURRENT);
  $trans_obj->postDownload($locale, $completed);
}

function _lingotek_notification_handle_interim_download($trans_obj, $document_id) {
  $locale = "";
  if (isset($_GET['locale'])) {
    $locale = $_GET['locale']; // Project Callbacks
  } else if (isset($_GET['target_code'])) {
    $locale = $_GET['target_code']; // Integration Callback
  }

  $completed = false;
  if ((isset($_GET['complete']) && $_GET['complete'] != 'false') || (isset($_GET['completed']) && $_GET['completed'] != 'false')){
    $completed = true;
  }
  $project_id = isset($_GET['project_id']) ? $_GET['project_id'] : "";
  $trans_obj->preDownload($locale, $completed);

  $replacements = array(
    '@trans_type' => get_class($trans_obj),
    '@document' => $document_id,
    '@language_code' => $locale,
    '@project_id' => $project_id,
    '@id' => $trans_obj->getId(),
  );

  if ($trans_obj->downloadTriggered($locale)) {
    LingotekLog::trace('Updated local content for <strong>@trans_type</strong> @id based on hit
          from external API for document: @document, language code @language_code, project ID: @project_id', $replacements, 'api');
  }
  else {
    LingotekLog::trace('Unable to update local content for <strong>@trans_type</strong> @id based on hit
          from external API for document: @document, language code @language_code, project ID: @project_id', $replacements, 'api');
  }
  $trans_obj->setTargetsStatus(LingotekSync::STATUS_INTERIM, $locale);
  $trans_obj->postDownload($locale, $completed);
}

function _lingotek_notification_handle_document_uploaded_or_updated($trans_obj) {
  $trans_obj->setStatus(LingotekSync::STATUS_CURRENT);
}

function _lingotek_notification_handle_import_failure($trans_obj, $document_id) {
  _lingotek_notification_handle_failed_upload($trans_obj, $document_id, "Source document import failed. Check TMS for details.");
}

function _lingotek_notification_handle_document_archived($trans_obj) {
  LingotekSync::disassociateEntitiesUsingEntityId($trans_obj->getId());
  $trans_obj->setTargetsStatus(LingotekSync::STATUS_ARCHIVED);
  $trans_obj->setStatus(LingotekSync::STATUS_ARCHIVED);
}

function _lingotek_notification_handle_document_deleted($trans_obj) {
  LingotekSync::disassociateEntitiesUsingEntityId($trans_obj->getId());
  $trans_obj->setTargetsStatus(LingotekSync::STATUS_GONE);
  $trans_obj->setStatus(LingotekSync::STATUS_GONE);
}

function _lingotek_notification_handle_target_deleted($trans_obj, $locale) {
  LingotekSync::setTargetStatus($trans_obj->getEntityType(), $trans_obj->getId(), $locale, LingotekSync::STATUS_GONE);
}



/**
 * Registers the site translation notfication callback.
 */
function lingotek_notifications() {
  drupal_page_is_cacheable(FALSE);
  LingotekLog::trace('Received <pre>@data</pre>', array('@data' => check_plain(var_export($_GET, TRUE))), 'callback');

  $callback_type = isset($_GET['type']) ? $_GET['type'] : 'phase_integration';
  $locale = isset($_GET['locale']) ? $_GET['locale'] : null;
  $document_id = NULL;
  if (isset($_GET['document_id'])) {
    $document_id = $_GET['document_id']; // Project Callback
  } else if (isset($_GET['doc_id'])) {
    $document_id = $_GET['doc_id']; // Integration Callback
  }

  $document_idx = NULL;
  if (isset($_GET['documentId'])) {
    $document_idx = $_GET['documentId']; // Project Callback
  } else if (isset($_GET['doc_idx'])) {
    $document_idx = $_GET['doc_idx']; // Integration Callback
  }

  LingotekLog::trace("Received callback of type $callback_type\n");
  if (!$document_id && !$document_idx) {
    return _lingotek_notification_output("No document id or document idx");
  }

  $trans_obj = lingotek_get_trans_obj($document_id, $document_idx);
  if (!$trans_obj) {
    return _lingotek_notification_output("Document not found. document_id " . $document_id . ", document_idx " . $document_idx);
  }

  switch ($callback_type) {
    case 'phase_integration':
    case 'target':
    case 'phase':
      LingotekLog::trace("Handling target or phase\n");
      _lingotek_notification_handle_target_or_phase($trans_obj, $document_id);
      break;
    case 'download_interim_translation':
      LingotekLog::trace("Handling interim download\n");
      _lingotek_notification_handle_interim_download($trans_obj, $document_id);
      break;
    case 'document_uploaded':
    case 'document_updated':
      LingotekLog::trace("Handling uploaded or updated\n");
      _lingotek_notification_handle_document_uploaded_or_updated($trans_obj);
      break;
    case 'import_failure':
      LingotekLog::trace("Handling import failure\n");
      _lingotek_notification_handle_import_failure($trans_obj, $document_id);
      break;
    case 'document_archived':
      LingotekLog::trace("Handling document archived\n");
      _lingotek_notification_handle_document_archived($trans_obj);
      break;
    case 'document_deleted':
      LingotekLog::trace("Handling document deleted\n");
      _lingotek_notification_handle_document_deleted($trans_obj);
      break;
    case 'target_deleted':
      LingotekLog::trace("Handling target deleted\n");
      _lingotek_notification_handle_target_deleted($trans_obj, $locale);
      break;
    default:
      LingotekLog::trace("Unknown callback\n");
      _lingotek_notification_output("Callback received, no handle for type: " . $callback_type);
  }
}

/**
 * The API endpoint for bulk translation management
 */
function lingotek_sync_endpoint() {
  $parameters = array();
  $method = $_SERVER['REQUEST_METHOD'];
  $status = "200";
  $request = array(
    'method' => $method,
  );
  $response = array();

  switch ($method) {
    case 'GET':
      $temp = array();
      foreach ($_GET as $key => $value) {
        $temp[$key] = check_plain($value);
      }
      $request['parameters'] = $parameters = $temp;
      /* $request['doc_ids'] = $document_ids = isset($parameters['doc_ids']) ? array_map(function($val) {
        return trim($val);
        }, explode(',', $parameters['doc_ids'])) : array(); */
      $response = LingotekSync::getReport();
      break;
    case 'POST': case 'PUT': case 'DELETE':
    default:
      parse_str(file_get_contents("php://input"), $parameters);
      $status = "405 Method Not Allowed";
      break;
  }

  return lingotek_json_output_cors($response, $status, array('methods_allowed' => 'GET'));
}

/**
* Updates the 'target_sync_status_[lang-code]' field for every target in the lingotek table
* with the overall progress returned by TMS
*
* @param int array $document_ids
*    array of Document IDs that you want to update
*
*/
function lingotek_get_and_update_target_progress(&$update_context = array(), $entity_type = NULL, $document_ids = NULL, $current_nids = -1, $total_nids = -1) {
  if (empty($update_context)) {
    return;
  }
  $api = LingotekApi::instance();
  if (isset($update_context['entity'])) {
    return check_entity_progress_for_download($update_context, $api);
  }
  if (empty($document_ids)) {
    return;
  }
  if (!is_array($document_ids)) {
    $document_ids = array($document_ids);
  }
  $failed_doc_ids = update_pending_sources($document_ids, $api);
  $update_context['message'] = t('Checking status of translations (@current of @total complete)', array('@current' => $current_nids, '@total' => $total_nids));
  $project_id = variable_get('lingotek_project', NULL);
  $progress_report = $api->getProgressReport($project_id, $document_ids);
  if (isset($progress_report) && $progress_report->results == 'success') {
    $delete_nids_maybe = array();
    $entity_values = array();

    if (count(get_object_vars($progress_report->errors)) > 0) {
      foreach (get_object_vars($progress_report->errors) as $doc_id => $error) {
        list($entity_id, $entity_type) = LingotekSync::getEntityIdFromDocId($doc_id, $entity_type);
        switch ($error->status) {
          case 'IN_QUEUE':
            LingotekSync::setUploadStatus($entity_type, $entity_id, LingotekSync::STATUS_PENDING);
            break;
          case 'NOT_FOUND':
          default:
            LingotekSync::setUploadStatus($entity_type, $entity_id, LingotekSync::STATUS_ERROR);
            lingotek_keystore($entity_type, $entity_id, 'last_sync_error', substr($error, 0, 255));
            LingotekLog::error('Received unexpected error status (@status) from Lingotek for @entity_type #@id: <pre>@error</pre>', array('@status' => $error->status, '@entity_type' => $entity_type, '@id' => $entity_id, '@error' => $error));
        }
      }
    }
    foreach ($progress_report->byDocumentIdAndTargetLocale as $doc_id => $completion) {
      list($entity_id, $entity_type_from_table) = LingotekSync::getEntityIdFromDocId($doc_id);
      if (!$entity_id) {
        LingotekLog::error("Lingotek doc ID '@doc_id' not found", array('@doc_id' => $doc_id));
        continue;
      }
      elseif (in_array($doc_id, $failed_doc_ids)) {
        LingotekLog::error("Doc ID '@doc_id' failed to update. Check TMS for details", array('@doc_id' => $doc_id));
        continue;
      }
      else {
        $delete_nids_maybe[] = $entity_id;
      }
      $entity_source_status = lingotek_keystore($entity_type, $entity_id, 'upload_status');

      foreach ($completion as $language => $percent) {
        $status = LingotekSync::getTargetStatus($doc_id, $language);
        $to_status = $status;
        if (isset($progress_report->workflowCompletedByDocumentIdAndTargetLocale->$doc_id->$language)) {
          if ($progress_report->workflowCompletedByDocumentIdAndTargetLocale->$doc_id->$language) { // If the workflow is complete
            if ($status != LingotekSync::STATUS_CURRENT) { // If the status is not current
              $to_status = LingotekSync::STATUS_READY; // Set it to ready
            }
            else {
              $to_status = LingotekSync::STATUS_CURRENT; // Otherwise keep it at current
            }
          }
          else { // If the workflow is not complete
            if ($percent == 0) {
              $to_status = LingotekSync::STATUS_PENDING; // Set it to pending
            }
            elseif ($percent > 0) {
              $to_status = LingotekSync::STATUS_READY_INTERIM;
            }
            else {
              $to_status = $status;
            }
          }
          if ($status != LingotekSync::STATUS_UNTRACKED) {
            $entity_values[] = array($entity_type, $entity_id, 'target_sync_status_' . $language, $to_status);
          }
        }
      }
      // update source status when necessary
      if ($entity_source_status == LingotekSync::STATUS_ERROR || $entity_source_status == LingotekSync::STATUS_PENDING) {
        LingotekSync::setUploadStatus($entity_type, $entity_id, LingotekSync::STATUS_CURRENT);
      }
    }
    // merge status info for entities
    foreach ($entity_values as $record) {
      $entity_id = (isset($record['entity_id']) ? $record['entity_id'] : $record[1]);
      $entity_key = (isset($record['entity_key']) ? $record['entity_key'] : $record[2]);
      $value = (isset($record['value']) ? $record['value'] : $record[3]);
      db_merge('lingotek_entity_metadata')
          ->key(array('entity_id' => $entity_id, 'entity_type' => $entity_type, 'entity_key' => $entity_key))
          ->fields(array(
            'entity_type' => $entity_type,
            'entity_id' => $entity_id,
            'entity_key' => $entity_key,
            'value' => $value,
          ))
          ->execute();
      lingotek_cache_clear($entity_type, $entity_id);
    }
    return $progress_report;
  }
  else {
    $error_message = t("API Error(s):") . " <ul>";
    if (is_array($progress_report->errors)) {
      foreach ($progress_report->errors as $error) {
        $error_message .= "<li>" . $error . "</li>";
      }
    }
    $error_message .= "</ul><i>" . t('For additional information, check your <a href="@link">recent log messages</a>', array('@link' => url('admin/reports/dblog'))) . "</i>";
    drupal_set_message(filter_xss($error_message), 'error');
  }
}

function _lingotek_notification_handle_failed_upload(LingotekEntity $ln, $doc_id, $error) {
  $ln->deleteMetadataValue("process_id");
  $ln->setStatus(LingotekSync::STATUS_ERROR);
  if ($ln->getMetadataValue("initial_upload")) {
    $ln->deleteMetadataValue("document_id");
    $ln->deleteMetadataValue("initial_upload");
  }
  else {
    $ln->setMetadataValue("document_id", $ln->getMetadataValue("previous_doc_id"));
  }
  $ln->setExclusiveTargetsStatus(LingotekSync::STATUS_NONE, [LingotekSync::STATUS_CURRENT, LingotekSync::STATUS_READY]);
  drupal_set_message(t('Document @id failed to update/upload. Check TMS for details. Given Error: @error', array('@id' => $doc_id, '@error' => $error)), 'error');
}

function update_pending_sources($document_ids, $api) {
  if (!is_array($document_ids)) {
    $document_ids = array($document_ids);
  }
  $failed_doc_ids = [];
  foreach ($document_ids as $doc_id) {
    $ln = LingotekEntity::loadByLingotekDocumentId($doc_id);
    if (!$ln) {
      continue;
    }
    $process_id = $ln->getMetadataValue('process_id');
    if (empty($process_id)) {
      continue;
    }
    $result = $api->sourceCompleted($process_id);
    if (!is_bool($result))  {
      $failed_doc_ids[] = $doc_id;
      _lingotek_notification_handle_failed_upload($ln, $doc_id, $result);
      continue;
    }
    if ($result) {
      $ln->setStatus(LingotekSync::STATUS_CURRENT);
      $ln->deleteMetadataValue("process_id");
      $ln->deleteMetadataValue("previous_doc_id");
      $ln->deleteMetadataValue("initial_upload");
    }
  }
  return $failed_doc_ids;
}

function check_entity_progress_for_download($update_context, $api) {
    $document_id = $update_context['entity']->lingotek['document_id'];
    $doc_failed = update_pending_sources($document_id, $api);
    if (!empty($doc_failed)) {
      return;
    }
    $project_id = variable_get('lingotek_project', NULL);
    $id = lingotek_entity_extract_ids($update_context['entity_type'], $update_context['entity']);
    if (empty($document_id) || empty($project_id) || empty($id)) {
      return LingotekLog::error("Issue updating target process on line @line in file @file", array('@line' => __LINE__, '@file' => __FILE__));
    }
    $progress_report = $api->getProgressReport($project_id, array($document_id));
    $protected_statuses = array(LingotekSync::STATUS_UNTRACKED, LingotekSync::STATUS_CURRENT, LingotekSync::STATUS_INTERIM);
    if (isset($progress_report) && isset($progress_report->byTargetLocale) && isset($progress_report->workflowCompletedByDocumentIdAndTargetLocale) && $progress_report->results == "success") {
      foreach ($progress_report->byTargetLocale as $locale => $percent_complete) {
        if (in_array(LingotekSync::getTargetStatus($document_id, $locale), $protected_statuses)) {
          continue;
        }
        elseif (!empty($progress_report->workflowCompletedByDocumentIdAndTargetLocale->$document_id->$locale)) {
          // There was an issue where the progress report was saying that the target was incomplete when it WAS complete
          lingotek_keystore($update_context['entity_type'], $id[0], 'target_sync_status_' . $locale, LingotekSync::STATUS_READY);
          continue;
        }

        if ($percent_complete == 0) {
          lingotek_keystore($update_context['entity_type'], $id[0], 'target_sync_status_' . $locale, LingotekSync::STATUS_PENDING);
        }
        elseif ($percent_complete > 0 && $percent_complete < 100) {
          lingotek_keystore($update_context['entity_type'], $id[0], 'target_sync_status_' . $locale, LingotekSync::STATUS_READY_INTERIM);
        }
        elseif ($percent_complete == 100) {
          lingotek_keystore($update_context['entity_type'], $id[0], 'target_sync_status_' . $locale, LingotekSync::STATUS_READY);
        }
        else {
          LingotekLog::error("Unknown status for locale @locale on nodeID #@current_nid.", array('@locale' => $locale, '@current_nid' => $id[0]));
        }
      }
    }
}


/**
* Updates the 'target_sync_status_[lang-code]' field for every target in the
* lingotek_config_metadata table with the overall progress returned by TMS
*
* @param int array $document_ids
*    array of Document IDs to be updated
*
*/
function lingotek_update_config_progress($document_ids) {
  $api = LingotekApi::Instance();
  if (empty($document_ids)) {
    return;
  }
  if (!is_array($document_ids)) {
    $document_ids = array($document_ids);
  }
  $config_profile = LingotekProfile::loadById(LingotekSync::PROFILE_CONFIG);
  $project_id = $config_profile->getProjectId();

  $progress_report = $api->getProgressReport($project_id, $document_ids);

  if (isset($progress_report) && $progress_report->results == 'success') {
    $cids = array();
    $cfg_values = array();
    $trans_obj = NULL;

    if (isset($progress_report->errors)) {
      foreach (get_object_vars($progress_report->errors) as $doc_id => $error) {
        $set = LingotekConfigSet::loadByLingotekDocumentId($doc_id);
        switch ($error->status) {
          case 'IN_QUEUE':
            $set->setMetadataValue('upload_status', LingotekSync::STATUS_PENDING);
            break;
          case 'NOT_FOUND':
          default:
            $set->setMetadataValue('upload_status', LingotekSync::STATUS_ERROR);
            LingotekLog::error('Received unexpected error status (@status) from Lingotek for config chunk #@id: <pre>@error</pre>', array('@status' => $error->status, '@id' => $doc_id, '@error' => $error));
        }
      }
    }

    foreach ($progress_report->byDocumentIdAndTargetLocale as $doc_id => $completion) {
      $trans_obj = LingotekConfigSet::loadByLingotekDocumentId($doc_id);
      if (!$trans_obj) {
        LingotekLog::error("Lingotek doc ID '@doc_id' not found", array('@doc_id' => $doc_id));
        continue;
      }

      foreach ($completion as $language => $percent) {
        $status = LingotekSync::getTargetStatus($doc_id, $language);
        if (isset($progress_report->workflowCompletedByDocumentIdAndTargetLocale->$doc_id->$language)) {
          if ($progress_report->workflowCompletedByDocumentIdAndTargetLocale->$doc_id->$language) { // If the workflow is complete
            if ($status != LingotekSync::STATUS_CURRENT) { // If the status is not current
              $to_status = LingotekSync::STATUS_READY; // Set it to ready
            }
            else {
              $to_status = LingotekSync::STATUS_CURRENT; // Otherwise keep it at current
            }
          }
          else { // If the workflow is not complete
            if ($percent == 0) {
              $to_status = LingotekSync::STATUS_PENDING; // Set it to pending
            }
            elseif ($percent > 0) {
              $to_status = LingotekSync::STATUS_READY_INTERIM;
            }
            else{
              $to_status = $status;
            }
          }
          $cfg_values[] = array($trans_obj->sid, 'target_sync_status_' . $language, $to_status);
        }
      }
    }

    // insert status info for config
    foreach ($cfg_values as $record) {
      $query = db_merge('lingotek_config_metadata')
          ->key(array('id' => $record[0], 'config_key' => $record[1]))
        ->fields(array(
            'id' => $record[0],
            'config_key' => $record[1],
            'value' => $record[2]
          ))
        ->execute();
    }
    return $progress_report;
  }
  else {
    $error_message = t("API Error(s):") . " <ul>";
    if (is_array($progress_report->errors)) {
      foreach ($progress_report->errors as $error) {
        $error_message .= "<li>" . $error . "</li>";
      }
    }
    $error_message .= "</ul><i>" . t('For additional information, check your <a href="@link">recent log messages</a>', array('@link' => url('admin/reports/dblog'))) . "</i>";
    drupal_set_message(filter_xss($error_message), 'error');
  }

}
